#include "chil_global.h"

#define atoui(c) atoi(atoui_char_p = (c)); if (atoui_char_p < 0) return -1
char *atoui_char_p;

static
char *
__CHILAppendCustom(char *orig, char *new, char separator)
{
  if (!orig) {
    /* orig is NULL, so allocate memory and copy new to it (unless new is NULL too)*/
    if (!new) {
      char *ret = CHILCalloc(1, sizeof(char));
      *ret = '\0';
      return ret;
    }
    else {
      int len = strlen(new) + 1;
      char *ret = CHILCalloc(len, sizeof(char));
      strcpy(ret, new);
      ret[len-1] = '\0';
      return ret;
    }
  }
  else {
    if (!new) {
      return orig;
    }
    else {
      int len_orig = strlen(orig);
      int len_new = strlen(new);
      char *ret = CHILRealloc(orig, len_orig + len_new + 2);
      strcpy(ret + len_orig + 1, new);
      /* Put a space or newline between the strings */
      ret[len_orig] = separator;
      ret[len_orig + len_new + 1] = '\0';
      return ret;      
    }
  }
}

#define __CHILAppend(orig, new) __CHILAppendCustom(orig, new, '\n')
#define __CHILAppendWithSpace(orig, new) __CHILAppendCustom(orig, new, ' ')

void
CHILUsage(ChilboltonData *data)
{
  char *program_name = "{program}";

  if (data->argv) {
    program_name = data->argv[0];
  }

  CHILPrintf("Usage:\n"
	     "   %s file1.dat raster1-rasterN [file2.dat raster1-rasterN ...]"
	     " [options]\n"
	     "   %s [-all|-raw] file1.{dat|raw} [file2.{dat|raw} ...] [options]\n"
	     "   %s -date YYYYMMDD filepattern1 [filepattern2 ...] [options]\n",
	     program_name, program_name, program_name);
  return;
}

void
CHILOptions(ChilboltonData *data)
{
  CHILPrintf("\n"
             "General data processing options (f=float n=integer s=string):\n"
	     "      -averays n  average n rays\n"
             "      -avetime n1 n2\n"
	     "                  average onto a regular n1-second grid but taking no less\n"
	     "                     then n2 rays in each n1 seconds\n"
	     "     -avegates n  average n gates\n"
	     "      -verbose    display raster and file information\n"
	     "        -quiet    display no information except errors\n"
	     "      -version    output version information\n"
	     "          -day n  only accept data written on day n\n"
	     "  -rangeoffset f  perform range correction with offset of f metres\n"
	     "         -Zcal f  add f dB to raw Zh values\n"
	     "      -autocal    use figures in summary file to calibrate Z and ZDR\n"
	     "        -clean    remove erroneous lines one pixel thick\n"
	     "  -doubleclean    remove erroneous lines one or two pixels thick\n"
	     "  -customclean n1 n2 n3\n"
             "                  remove isolated groups of n3 or fewer cloudy gates\n"
             "                     every n1 gates, offset by n2\n"
	     "  -interpclean n1 n2 n3\n"
             "                  interpolate over groups of n3 bad gates\n"
             "                     every n1 gates, offset by n2\n"
	     "        -vfold f  set the folding velocity for all velocities to\n"
	     "                     +/- f m/s\n"
	     "       -v1fold f  set the folding velocity of v1\n"
	     "       -v2fold f  set the folding velocity of v2\n"
	     "     -reversev    negate all velocities\n"
	     "      -minelev f  skip rays with an elevation less than f degrees\n"
	     "    -maxfields n  retrieve only the first n fields\n"
	     "   -extragates n  retrieve n extra gates in data with incorrect raster header\n"
	     " -velocityaway    positive velocities are away from the radar\n"
	     "       -sumdir s  set fallback summary-file directory to s\n"
	     "                     (default is current directory)\n"
	     "\n"
	     "Processing options specific to scanning data:\n"
	     "         -scan n  get scan n\n"
	     "       -ZDRcal f  add f dB to raw ZDR values\n"
	     "       -LDRcal f  add f dB to raw LDR values\n"
	     "     -PHIDPcal f  add f degrees to raw PHIDP values\n"
	     "       -maxLDR f  remove all data (probably clutter) with LDR greater than f\n"
	     "                     (-10 is a good value)\n"
	     "        -noise f  subtract noise from Z using noise-equivalent reflectivity at\n"
	     "                     1 km of f dBZ\n"
	     "        -polar fZH fZV fZHV\n"
	     "                  subtract noise from Z, ZDR and LDR using noise-equivalent\n"
	     "                     reflectivities at 1 km in the H, V and cross-polar\n"
	     "                     channels of fZ, fZV and fZHV (dBZ) respectively\n"
	     "    -threshold f  set noise-elemination threshold to f\n"
	     "      -doppler f  reject noisy velocity and phi_dp data with threshold f\n"
	     "\n"
	     "Processing options specific to cloud data:\n"
	     "      -process    subtract noise (Zh must be present) using gates\n"
	     "                  from the top of the ray to characterise noise\n"
	     "   -oscillator    subtract noise using gates from the top of the ray\n"
	     "                     to characterise oscillator interference\n"
	     "   -noisegates n  use n gates to characterise noise/interference\n"
	     "  -noisefactor f  set noise-elimination factor to f\n"
	     "    -noclutter n f   suppress clutter by removing data when both v and sigma_v\n"
	     "                     are less than f in the lowest n gates\n"
	     "      -nobreak    don't break data up into days\n"
	     " -correctrange    perform range correction\n"
	     "      -v_floor f  set equivalent noise floor for v to f dB above that for Zh\n"
	     "     -sv_floor f  set equiv. noise floor for sigma_v to f dB above that for Zh\n"
	     "       -reject n f   reject `range-glitched' rays with a Zh that jumps by more\n"
	     "                     than f in two gates beyond gate n\n"
	     " -cluttergates n  if there are fewer than n clutter/transmit-pulse gates, reject ray\n"
	     "     -pciclean    skip rays that have been contaminated on the PCI system\n");

  return;
}


int
CHILInit(int argc, char **argv, ChilboltonData *data)
{
  int i = 1, j, datelen, status;

  data->firsttime = 1;
  data->scan = 0;
  data->avegates = data->averays = 1;
  data->minrays = 1;
  data->ongrid = 0;
  data->process = data->clean = 0;
  data->file = 0;
  data->noisegates = 25;
  data->newrangeoffset = data->noclutter = data->warn_glitch = 0;
  data->polarprocess = data->dopplerprocess = 0;
  data->warn_cloudnoise = data->warn_polarnoise = data->warn_low_sensitivity = 0;
  data->warn_clutterLdr = data->warn_oscillator = data->warn_minelev = 0;
  data->warn_confused = data->warn_dopplernoise = data->warn_calchange = 0;
  data->nobreak = 0;
  data->filebuf = NULL;
  data->threshold = 1.19;
  data->dopplerthreshold = 400.0;
  data->iray = data->iblock = data->nblock = 0;
  data->noisefactor = 3.0;
  data->rangeoffset = 0.0;
  data->Zcal = data->ZDRcal = data->PHIDPcal = data->LDRcal = 0.0;
  data->nez_Zh = data->nez_Zv = data->nez_Zhv = -99;
  data->maxLDR = 101;
  data->optstart = 0;
  data->maxfields = 127;
  data->extragates = 0;
  data->verbose = data->quiet = data->progress = 0;
  data->reqday = data->reqmonth = data->reqyear = 0;
  data->rangecorrection = 0;
  data->sigma_v_floor = data->v_floor = 0.0;
  data->reject = 0;
  data->reject_gate = 0;
  data->reject_val = 0;
  data->minelev = -1000;
  data->noclutter_minv = 0.1;
  data->noclutter_maxgate = 15;
  data->v1fold = data->v2fold = -1.0;
  data->reversev = 0;
  data->custom_clean = 0;
  data->interp_cycle = data->clean_cycle = data->noisegates;
  data->clean_offset = 0;
  data->clean_gates = 1;
  data->interp_clean = 0;
  data->interp_offset = 0;
  data->interp_gates = 1;
  data->pci_clean = data->min_clutter_gates = 0;
  data->autocal = data->velocity_away = 0;
  data->forceZcal = data->forceZDRcal = data->forceLDRcal = data->forcePHIDPcal = 0;
  data->use_v1fold = data->use_v2fold = 0;

  data->argv = argv;
  data->argc = argc;

  /* Reset some things to avoid segfaults if CHILEnd is called... */
  data->chil_file.position = FILE_CLOSED;
  data->chil_file.file = NULL;
  data->ray.range = NULL;
  data->file_list.allnames = NULL;

  if (argc < 2) {
    return CHIL_NOTFOUND;
  }
  if (strcmp(argv[1], "-date") == 0) {
    /* The program must read all summary files to decide what rasters
       to load */
    if (argc < 3) {
      CHILPrintf("No date specified!");
      return CHIL_ERROR;
    }
    datelen=strlen(argv[2]);
    if (datelen==6) {
      i=sscanf(argv[2],"%02u%02u%02u",
	       &data->reqyear, &data->reqmonth, &data->reqday);
    }
    else if (datelen==8) {
      i=sscanf(argv[2],"%04u%02u%02u",
	       &data->reqyear, &data->reqmonth, &data->reqday);
      data->reqyear -= 1900;
    }
    if (i<3) {
      CHILPrintf("Can't understand date: should be in the form `yymmdd' or `yyyymmdd'\n");
      return CHIL_ERROR;
    }
    if (argc < 4) {
      CHILPrintf("Need at least one filename\n");
      return CHIL_ERROR;
    }

    /* Must look for "-sumdir" argument! */
    for (i = 4; i < argc; ++i) {
      if (strcmp(argv[i],"-sumdir") == 0 && i+1<argc) {
	CHILSecondarySumDir(argv[++i]);
      }
    }

    if (CHILGetFileList(data->reqday, data->reqmonth, data->reqyear,
			argc-3, argv+3,
			&data->file_list) < 1) {
      CHILPrintf("Can't find any data in these files from %04u%02u%02u\n",
		 CHILTrueYear(data->reqyear), data->reqmonth, data->reqday);
      return CHIL_NOTFOUND;
    }
    /* Find the first switch */
    for (j = 4; j < argc; j++) {
      if (*argv[j] == '-' && *(argv[j]+1) != '\0')
	break;
    }
  }
  else if (strcmp(argv[1], "-all") == 0 || strcmp(argv[1], "-raw") == 0) {
    /* We want all rasters in all the files specified on the command line */
    if (argc < 3 || *argv[2] == '-') {
      CHILPrintf("No files specified\n");
      return CHIL_ERROR;
    }
    data->file_list.allnames = argv+2;
    /* Count the files / find the first switch */
    for (j = 3; j < argc && *argv[j] != '-'; j++);
    data->file_list.n = j-2;
  }
  else {
    /* The first pairs of arguments are filenames and raster ranges.
       For non-cloud data one would usually expect a single file and raster */
    j = 1;
    data->file_list.n = 0;
    while (j < argc) {
      /* first the filename */
      data->file_list.name[data->file_list.n] = argv[j];

      if (*(data->file_list.name[data->file_list.n]) == '-')
	/* found a switch: end of file/raster pairs */
	break;
      /* then the raster range */
      if (++j >= argc) {
	CHILPrintf("No raster number(s) specified for file\n");
	return CHIL_ERROR;
      }
      if (*argv[j] == '-' && *(argv[j]+1) == '\0') {
	data->file_list.allrasters[data->file_list.n] = 1;
      }
      else {
	data->file_list.allrasters[data->file_list.n] = 0;
	i = sscanf(argv[j],"%hu-%hu",&data->file_list.first[data->file_list.n],
		   &data->file_list.last[data->file_list.n]);
	if (i == 0) {
	  CHILPrintf("Cannot understand `%s' as raster range: reading all rasters\n",
		     argv[j]);
	  data->file_list.allrasters[data->file_list.n] = 1;
	} else if (i == 1) {
	  data->file_list.last[data->file_list.n]
	    = data->file_list.first[data->file_list.n];
	} else if (data->file_list.last[data->file_list.n] 
		   < data->file_list.first[data->file_list.n]) {
	  CHILPrintf("Last raster number is less than first raster number!\n");
	  data->file_list.last[data->file_list.n] 
	    = data->file_list.first[data->file_list.n];
	}
      }
      j++;
      (data->file_list.n)++;
    }
  }
  data->optstart = j;
  if (j < argc) {
    /* we have some switches to parse */
    for (i = j; i < argc; i++) {
      if (strcmp(argv[i],"-averays") == 0 && i+1<argc) {
	data->averays = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-avetime") == 0 && i+2<argc) {
	data->averays = atoui(argv[++i]);
        data->minrays = atoui(argv[++i]);
        data->ongrid = 1;
      } else if (strcmp(argv[i],"-verbose") == 0) {
	data->verbose = 1;
      } else if (strcmp(argv[i],"-version") == 0) {
	CHILPrintf("Version: " CHIL_VERSION "\n");
	return CHIL_NOTFOUND;
      } else if (strcmp(argv[i],"-quiet") == 0) {
	data->quiet = 1;
      } else if (strcmp(argv[i],"-process") == 0) {
	data->process = 1;
      } else if (strcmp(argv[i],"-oscillator") == 0) {
	data->process = 2;
      } else if (strcmp(argv[i],"-noisegates") == 0 && i+1<argc) {
	data->noisegates = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-maxfields") == 0 && i+1<argc) {
	data->maxfields = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-extragates") == 0 && i+1<argc) {
	data->extragates = atoi(argv[++i]);
      } else if (strcmp(argv[i],"-clean") == 0) {
	/* If already -doubleclean, don't want to set data->clean to 1 */
	if (!data->clean) {
	  data->clean = 1;
	}
      } else if (strcmp(argv[i],"-doubleclean") == 0) {
	data->clean = 2;
      } else if (strcmp(argv[i],"-customclean") == 0 && i+2<argc) {
	data->clean_cycle = atoui(argv[++i]);
	data->clean_offset = atoui(argv[++i]);
	data->clean_gates = atoui(argv[++i]);
	if (data->clean_cycle <= 0 || data->clean_gates <= 0) {
	  data->custom_clean = 0;	  
	}
	else {
	  data->custom_clean = 1;
	}
      } else if (strcmp(argv[i],"-interpclean") == 0 && i+2<argc) {
	data->interp_cycle = atoui(argv[++i]);
	data->interp_offset = atoui(argv[++i]);
	data->interp_gates = atoui(argv[++i]);
	if (data->interp_cycle <= 0 || data->interp_gates <= 0) {
	  data->interp_clean = 0;	  
	}
	else {
	  data->interp_clean = 1;
	}
      } else if (strcmp(argv[i],"-velocityaway") == 0) {
	data->velocity_away = 1;
      } else if (strcmp(argv[i],"-pciclean") == 0) {
	data->pci_clean = 2;
      } else if (strcmp(argv[i],"-cluttergates") == 0 && i+1<argc) {
	data->min_clutter_gates = atoi(argv[++i]);
      } else if (strcmp(argv[i],"-noclutter") == 0 && i+2<argc) {
	data->noclutter = 1;
	data->noclutter_maxgate = atoui(argv[++i]);
	data->noclutter_minv = atof(argv[++i]);
      } else if (strcmp(argv[i],"-nobreak") == 0) {
	data->nobreak = 1;
      } else if (strcmp(argv[i],"-correctrange") == 0) {
	data->rangecorrection = 1;
      } else if (strcmp(argv[i],"-rangeoffset") == 0 && i+1<argc) {
	data->newrangeoffset = 1;
	data->rangecorrection = 1;
	data->rangeoffset = atof(argv[++i]);
      } else if (strcmp(argv[i],"-Zcal") == 0 && i+1<argc) {
	data->Zcal = atof(argv[++i]);
	data->forceZcal = 1;
      } else if (strcmp(argv[i],"-ZDRcal") == 0 && i+1<argc) {
	data->ZDRcal = atof(argv[++i]);
 	data->forceZDRcal = 1;
      } else if (strcmp(argv[i],"-LDRcal") == 0 && i+1<argc) {
	data->LDRcal = atof(argv[++i]);
	data->forceLDRcal = 1;
      } else if (strcmp(argv[i],"-PHIDPcal") == 0 && i+1<argc) {
	data->PHIDPcal = atof(argv[++i]);
	data->forcePHIDPcal = 1;
      } else if (strcmp(argv[i],"-autocal") == 0) {
	data->autocal = 1;
      } else if (strcmp(argv[i],"-threshold") == 0 && i+1<argc) {
	data->threshold = atof(argv[++i]);
      } else if (strcmp(argv[i],"-maxLDR") == 0 && i+1<argc) {
	data->maxLDR = atof(argv[++i]);
      } else if (strcmp(argv[i],"-polar") == 0 && i+3<argc) {
	data->nez_Zh = atof(argv[++i]);
	data->nez_Zv = atof(argv[++i]);
	data->nez_Zhv = atof(argv[++i]);
	data->polarprocess = 1;
      } else if (strcmp(argv[i],"-noise") == 0 && i+1<argc) {
	data->nez_Zh = atof(argv[++i]);
	data->polarprocess = 1;
      } else if (strcmp(argv[i],"-doppler") == 0 && i+1<argc) {
	data->dopplerthreshold = atof(argv[++i]);
	data->dopplerprocess = 1;
      } else if (strcmp(argv[i],"-reject") == 0 && i+2<argc) {
	data->reject_gate = atof(argv[++i]);
	data->reject_val = atof(argv[++i]);
	data->reject = 1;
      } else if (strcmp(argv[i],"-sumdir") == 0 && i+1<argc) {
	CHILSecondarySumDir(argv[++i]);
      } else if (strcmp(argv[i],"-noisefactor") == 0 && i+1<argc) {
	data->noisefactor = atof(argv[++i]);
      } else if (strcmp(argv[i],"-v_floor") == 0 && i+1<argc) {
	data->v_floor = atof(argv[++i]);
      } else if (strcmp(argv[i],"-sv_floor") == 0 && i+1<argc) {
	data->sigma_v_floor = atof(argv[++i]);
      } else if (strcmp(argv[i],"-vfold") == 0 && i+1<argc) {
	float vfold = atof(argv[++i]);
	if (vfold > 0) {
	  data->v1fold = vfold;
	  data->use_v1fold = 1;
	  data->v2fold = vfold;
	  data->use_v2fold = 1;
	}
	else {
	  CHILPrintf("Folding velocity is negative or zero: ignoring\n");
	}
      } else if (strcmp(argv[i],"-v1fold") == 0 && i+1<argc) {
	float vfold = atof(argv[++i]);
	if (vfold > 0) {
	  data->v1fold = vfold;
	  data->use_v1fold = 1;
	}
	else {
	  CHILPrintf("Folding velocity is negative or zero: ignoring\n");
	}
      } else if (strcmp(argv[i],"-v2fold") == 0 && i+1<argc) {
	float vfold = atof(argv[++i]);
	if (vfold > 0) {
	  data->v2fold = vfold;
	  data->use_v2fold = 1;
	}
	else {
	  CHILPrintf("Folding velocity is negative or zero: ignoring\n");
	}
      } else if (strcmp(argv[i],"-reversev") == 0) {
	data->reversev = 1;
      } else if (strcmp(argv[i],"-minelev") == 0 && i+1<argc) {
	data->minelev = atof(argv[++i]);
      } else if (strcmp(argv[i],"-avegates") == 0 && i+1<argc) {
	data->avegates = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-scan") == 0 && i+1<argc) {
	data->scan = atoui(argv[++i]);
      } else if (strcmp(argv[i],"-day") == 0 && i+1<argc) {
	data->reqday = atoui(argv[++i]);
      }
    }
  }

  if (data->file_list.n == 0) {
    CHILPrintf("No files specified!\n");
    return CHIL_ERROR;
  }

  if ((data->file_list.n>1 
       || data->file_list.first[0] < data->file_list.last[0])
      && (!data->quiet) && (!data->verbose))
    data->progress=1;

  data->file = -1;
    
  data->chil_file.position = FILE_CLOSED;
  status = CHILAdvance(data);
  if (status == CHIL_OK || status == CHIL_ENDOFBLOCK) {
    return CHIL_OK;
  }
  else {
    return status;
  }
}

int
CHILAdvance(ChilboltonData *data)
{
  char *file_name = "requested file";
  int status = 0;
  char allrasters = 1;
  ChilboltonFile *chil_file = &(data->chil_file);
  RasterInfo *raster_info = &(data->raster_info);
  Ray *ray = &(data->ray);

  if (data->file >= 0) {
    if (data->file_list.allnames) {
      allrasters = 1;
    }
    else {
      allrasters = data->file_list.allrasters[data->file];
    }
    if (chil_file->iscan == 0 || data->scan > 0) {
      ++(data->raster);
    }
  }
  while (1) {
    if ((data->file < 0)
	|| (chil_file->position == FILE_CLOSED)
	|| (chil_file->position == END_OF_DATA)
	|| (!allrasters && data->raster > data->file_list.last[data->file])) {

      if (chil_file->position != FILE_CLOSED)
	CHILClose(chil_file);

      if (++(data->file) < data->file_list.n) {
	if (data->file_list.allnames) {
	  file_name = data->file_list.allnames[data->file];
	}
	else {
	  file_name = data->file_list.name[data->file];
	}
	if ((status = CHILOpen(file_name, chil_file))
	    != START_OF_FILE) {
	  CHILPrintf("Failed to read %s\n", file_name);
	  return CHIL_NOTFOUND;
	}

	/* Set the calibration figures if required */
	if (data->autocal) {
	  if (chil_file->cals_constant) {
	    if (!data->forceZcal) {
	      if (!data->firsttime && data->Zcal != chil_file->Zcal) {
		data->warn_calchange = 1;
	      }
	      data->Zcal = chil_file->Zcal;
	    }
	    if (!data->forceZDRcal) {
	      if (!data->firsttime && data->ZDRcal != chil_file->ZDRcal) {
		data->warn_calchange = 1;
	      }
	      data->ZDRcal = chil_file->ZDRcal;
	    }
	  }
	  else if (chil_file->summary_exists) {
	    data->warn_calchange = 1;
	  }
	}
      }
      else {
	/*
	if (!data->quiet) { 
	  CHILPrintf("\n"); 
	} 
	*/
	return CHIL_ENDOFDATA;
      }

      if (data->file_list.allnames) {
	allrasters = 1;
      }
      else {
	allrasters = data->file_list.allrasters[data->file];
      }

      if (chil_file->summary_exists == 0 && !allrasters) {
	CHILClose(chil_file);
	CHILPrintf("Can't find summary file\n");
	return CHIL_NOTFOUND;
      }

      if (data->verbose)
	CHILPrintFileInformation(chil_file);

      if (data->progress)
	CHILPrintf("Opened %s\n", file_name);

      /*
      if (data->progress && !allrasters)
	CHILPrintf("   reading raster ");
      */

      if (!allrasters) {
	data->raster = data->file_list.first[data->file];
      }
    }

    if (((!allrasters) && chil_file->iscan == 0) || (data->scan > 0)) {

      if (data->progress)
	CHILPrintf("Reading raster %u\n", data->raster);

      if (CHILSeekRaster(chil_file, data->raster) == CHIL_OK ) {
	if (data->scan > 1) {
	  if (CHILSkipScans(chil_file, raster_info, data->scan-1)
	      != CHIL_OK) {
	    CHILPrintf("Warning: Scan %hu not found in raster\n", data->scan);
	    return CHIL_NOTFOUND;
	  }
	}
	break;
      }
      else {
	if (!data->quiet) {
	  CHILPrintf("Warning: Raster %u not found in %s\n", data->raster,
		     file_name);
	}
	++(data->raster);
	continue;
      }
    }
    break;
  }

  /* Found a good scan! */
  status = CHILReadInformationRecord(chil_file, raster_info);

  /* In cloud data recorded by the DSP system the max_gate is often 1 too low */
  raster_info->max_gate += data->extragates;

  if (status == END_OF_DATA && chil_file->position != END_OF_DATA) {
    /* Oh dear - don't know where we are in this file/raster (corrupt?)
       ... best to set the position to END_OF_DATA and so skip it */
    data->warn_confused = 1;
    chil_file->position = END_OF_DATA;
  }

  if (data->newrangeoffset) {
    raster_info->range_offset = data->rangeoffset;
  }

  if (data->verbose)
    CHILPrintRasterInformation(raster_info);

  if (!data->firsttime)
    CHILFreeRay(ray);

  CHILInitRay(raster_info, data->avegates, data->averays, ray);
  data->firsttime = 0;

  CHILSetFoldingVelocities(ray, data->v1fold, data->v2fold);

  ray->pci_clean = data->pci_clean;
  ray->minrays = data->minrays;

  ray->v_floor = data->v_floor;
  ray->sigma_v_floor = data->sigma_v_floor;

  if (data->verbose)
    CHILPrintFieldInformation(ray);

  if (raster_info->cloud_data) {
    return CHIL_OK;
  }
  else {
    /*
    if (!data->quiet) { 
      CHILPrintf("\n"); 
    } 
    */
    return CHIL_ENDOFBLOCK;
  }
}


int
CHILFetchRay(ChilboltonData *data)
{
  static int last_day = 0;
  static int last_month = 0;
  static int last_year = -1;
  int status;
  ChilboltonFile *chil_file = &(data->chil_file);
  RasterInfo *raster_info = &(data->raster_info);
  Ray *ray = &(data->ray);

  while (1) {
    if (data->ongrid) {
      status = CHILReadRayGrid(chil_file,raster_info,ray);
    }
    else {
      status = CHILReadRay(chil_file,raster_info,ray);
    }

    if (status == CHIL_OK) {

      if (data->reqday == 0 || ray->day == data->reqday) {

	if (raster_info->cloud_data) {
	  /* Cloud data */
	  if (ray->elevation < data->minelev) {
                printf("ray->elevation %12.6f data->minelev %12.6f\n",
                 	ray->elevation, data->minelev);

	    data->warn_minelev = 1;
	    continue;
	  }

	  if (data->reject) {
	    if (CHILIsRangeGlitched(ray, data->reject_gate, data->reject_val)
		== CHIL_ERROR) {
	      data->warn_glitch = 1;
	      continue;
	    }
	  }
	  if (data->process == 1) {
            if (CHILSubtractNoise(ray,data->noisegates,data->noisefactor)
		!= CHIL_OK)
	      data->warn_cloudnoise = 1;
	  }
	  else if (data->process == 2) {
            if (CHILSubtractOscillatorNoise(ray,data->noisegates,
					    data->noisefactor)
		!= CHIL_OK)
	      data->warn_oscillator = 1; 
	  }

	  if (data->min_clutter_gates) {
	    if (CHILIsLowSensitivity(ray, data->min_clutter_gates)) {
	      data->warn_low_sensitivity = 1;
	      continue;
	    }
	  }

	  if (data->noclutter)
            CHILSuppressClutter(ray, data->noclutter_maxgate,
				data->noclutter_minv);

	  if (data->clean == 1)
            CHILCleanRay(ray);
	  else if (data->clean == 2)
	    CHILDoubleCleanRay(ray);

	  if (data->interp_clean)
	    CHILInterpolateRay(ray, data->interp_cycle,
			       data->interp_offset, data->interp_gates);
	  else if (data->custom_clean)
	    CHILCustomCleanRay(ray, data->clean_cycle,
			       data->clean_offset, data->clean_gates);

	  if (data->rangecorrection)
	    CHILRangeCorrection(raster_info,ray);

	  if (data->Zcal != 0.0)
	    CHILCalibrateZh(ray,data->Zcal);

	  if (data->reversev)
	    CHILReverseVelocities(ray);

	  if (!data->nobreak && last_day > 0 
	      && (last_day != ray->day || last_month != ray->month 
		  || !(last_year == ray->year 
		       || last_year == ray->year + 100))) {
	    status = CHIL_ENDOFBLOCK;
	    
	    /*
	    if (!data->quiet) { 
	      CHILPrintf("\n"); 
	    } 
	    */
	  }
	  else {
	    status = CHIL_OK;
	    /*
	    if (!data->quiet) {
	      CHILPrintf(".");
	    }
	    */
	  }

	  last_day = ray->day;
	  last_month = ray->month;
	  last_year = ray->year;

	  return status;

	}
	else /* Scanning data */ {

	  if (ray->elevation < data->minelev) {
	    data->warn_minelev = 1;
	    continue;
	  }

	  if (data->polarprocess) {
            if (CHILSubtractNoisePolar(ray, data->nez_Zh, data->nez_Zv,
				       data->nez_Zhv,
				       data->threshold) != CHIL_OK)
	      data->warn_polarnoise = 1;
	  }
	  if (data->maxLDR < 100) {
            if (CHILSuppressClutterWithLdr(ray, data->maxLDR) != CHIL_OK)
	      data->warn_clutterLdr = 1;
	  }
	  if (data->dopplerprocess) {
	    if (CHILRemoveNoiseDoppler(ray, data->dopplerthreshold) != CHIL_OK)
	      data->warn_dopplernoise = 1;
	    /*
	    if (CHILDopplerClean(ray) != CHIL_OK)
	      data->warn_dopplernoise = 1;
	    */
	  }

	  if (data->clean == 1)
            CHILCleanRay(ray);
	  else if (data->clean == 2)
	    CHILDoubleCleanRay(ray);

	  if (data->interp_clean)
	    CHILInterpolateRay(ray, data->interp_cycle,
			       data->interp_offset, data->interp_gates);
	  else if (data->custom_clean)
	    CHILCustomCleanRay(ray, data->clean_cycle,
			       data->clean_offset, data->clean_gates);

	  if (data->rangecorrection) {
            CHILRangeCorrection(raster_info,ray);
	  }

	  if (data->Zcal != 0.0) 
	    CHILCalibrateZh(ray, data->Zcal);

	  if (data->ZDRcal != 0.0) 
	    CHILCalibrateField(ray, ZDR, data->ZDRcal);

	  if (data->LDRcal != 0.0) 
	    CHILCalibrateField(ray, LDR, data->LDRcal);

	  if (data->PHIDPcal != 0.0) 
	    CHILCalibrateField(ray, PHIDP, data->PHIDPcal);

	  if (data->reversev)
	    CHILReverseVelocities(ray);

	}

	/* Got a good ray and processed it */
	/*
	if (!data->quiet) { 
	  CHILPrintf("."); 
	} 
	*/
	return CHIL_OK;
      }

    }
    else /* CHILReadRay failed */ {
      if ((status = CHILAdvance(data)) != CHIL_OK) {
	return status;
      }
    }
  }
}

void
CHILEnd(ChilboltonData *data)
{
  if (!data->quiet) {
    if (data->warn_cloudnoise)
      CHILPrintf("Warning: Attempt to subtract noise failed\n");
    if (data->warn_polarnoise)
      CHILPrintf("Warning: Attempt to subtract noise from polarisation parameters failed\n");
    if (data->warn_dopplernoise)
      CHILPrintf("Warning: Attempt to remove noise from doppler parameters failed\n");
    if (data->warn_oscillator)
      CHILPrintf("Warning: Attempt to subtract oscillator noise failed\n");
    if (data->warn_clutterLdr)
      CHILPrintf("Warning: Attempt to suppress clutter with Ldr failed\n");
    if (data->warn_glitch)
      CHILPrintf("Warning: Some data was rejected because of range-glitching\n");
    if (data->warn_minelev)
      CHILPrintf("Warning: Some data was rejected because the radar elevation was less than %f degrees\n",
		 data->minelev);
    if (data->warn_confused)
      CHILPrintf("Warning: Data file appears to be truncated\n");
    if (data->warn_calchange)
      CHILPrintf("Warning: Z/ZDR calibration figures in summary files not constant so not used\n");
    if (data->warn_low_sensitivity)
      CHILPrintf("Warning: Some data was rejected because number of clutter gates suggested very low radar sensitivty\n");
  }
  CHILFreeRay(&(data->ray));
  if (data->chil_file.position != FILE_CLOSED) {
    CHILClose(&(data->chil_file));
  }
}

/*
int
CHILPrintOptions(FILE *file)
{
  int i, j;
  if (chil_argv == NULL)
    {
      return CHIL_NOTFOUND;
    }
  for (i = chil_optstart; i < chil_argc; i++)
    {
      char *c = chil_argv[i];
      for (j = 0; c[j] != '\0'; j++)
	{
	  if (c[j] <= ' ')
	    {
	      fprintf(file," \"%s\"", c);
	      break;
	    }
	}
      if (c[j] == '\0')
	{
	  fprintf(file," %s", c);
	}
    }
  return CHIL_OK;
}
*/

char *
CHILGetOptions(ChilboltonData *data)
{
  int i, j, len = 0;
  char *ans;

  if (data->argv == NULL)
    {
      return NULL;
    }
  for (i = data->optstart; i < data->argc; i++)
    {
      char *c = data->argv[i];
      char need_quotes = 0;
      for (j = 0; c[j] != '\0'; j++)
	{
	  if (c[j] <= ' ')
	    {
	      need_quotes = 1;
	    }
	}
      len += (1 + j + need_quotes * 2);
    }
  if (!len) {
    len = 1;
  }

  ans = CHILCalloc(len, sizeof(char));
  if (ans == NULL) {
    CHILPrintf("Out of memory\n");
    return NULL;
  }

  len = 0;
  for (i = data->optstart; i < data->argc; i++)
    {
      char *c = data->argv[i];
      char need_quotes = 0;
      for (j = 0; c[j] != '\0'; j++)
	{
	  if (c[j] <= ' ')
	    {
	      need_quotes = 1;
	    }
	}
      if (need_quotes) {
	sprintf(ans + len, "\"%s\"", c);
      }
      else {
	sprintf(ans + len, "%s", c);
      }
      len += (1 + j + need_quotes * 2);
      ans[len-1] = ' ';
    }
  if (len) {
    ans[len-1] = '\0';
  }
  else {
    ans[0] = '\0';
  }
  return ans;
}

#define NUMSTRLEN 20
/* The information reported by this function a dynamically allocated
 * string that should be freed with CHILFree() - the information is
 * held in the `comments' or `comment' attribute of NetCDF variables
 * written with chil2nc. */
char *
CHILFileInformation(ChilboltonData *data)
{
  char *inf = NULL;
  char numstring[NUMSTRLEN];
  
  if (data->pci_clean) {
    inf = __CHILAppend(inf, "The coherent processing system sometimes produces rays that are corrupted at low levels:\n"
		       "these rays have been rejected, so the remaining rays may not be evenly spaced in time.");
  }
  if (CHILFindField(&(data->ray), LDR) >= 0) {
    if (data->maxLDR <= 100) {
      snprintf(numstring, NUMSTRLEN, "%g", data->maxLDR);
      numstring[NUMSTRLEN-1] = '\0';
      inf = __CHILAppend(inf, "To reduce ground clutter, data has been removed (from all variables) whenever Ldr exceeds");
      inf = __CHILAppendWithSpace(inf, numstring);
      inf = __CHILAppendWithSpace(inf, "dB.");
    }
    else {
      inf = __CHILAppend(inf, "To reduce the effects of ground clutter the user should remove pixels"
			 " (from all variables) that have Ldr greater than around -10 dB.");
    }
  }
  if (!data->raster_info.cloud_data) {
    inf = __CHILAppend(inf, "In calculating the height of each data point, the curvature of the earth should be accounted for.");
  }
  return inf;
}

/* The information reported by this function a dynamically allocated
 * string that should be freed with CHILFree() - the information is
 * held in the global `comments' or `comment' attribute of NetCDF
 * files written with chil2nc. It should not duplicate information
 * reported by CHILFileInformation. */
char *
CHILFieldInformation(const ChilboltonData *data, int field)
{
  char *inf = NULL;
  char numstring[NUMSTRLEN];
  uchar f = data->ray.field_type[field];

  if (f == V1 || f == V2) {
    if (data->velocity_away) {
      inf = __CHILAppend(inf, "This parameter is the radial component of the velocity, with positive velocities away from the radar.");
    }
    else {
      inf = __CHILAppend(inf, "This parameter is the radial component of the velocity, with positive velocities towards the radar.");
    }
  }
  else if (f == WIDTH_H || f == WIDTH_V) {
    inf = __CHILAppend(inf, "This parameter is the standard deviation of the reflectivity-weighted velocities in the radar pulse volume.");
  }
  else if (f == V_STD) {
    snprintf(numstring, NUMSTRLEN, "%d", data->averays * data->avegates);
    numstring[NUMSTRLEN-1] = '\0';
    inf = __CHILAppend(inf, "The data in this file are at a lower resolution than the raw data, and this parameter is the standard deviation\n"
		       "of the");
    if (~data->ongrid) {
      inf = __CHILAppendWithSpace(inf, numstring);
    }
    inf = __CHILAppendWithSpace(inf, "raw Doppler velocities measured during in each output gate and ray.");
  }  
  else if (f == Z_STD) {
    snprintf(numstring, NUMSTRLEN, "%d", data->averays * data->avegates);
    numstring[NUMSTRLEN-1] = '\0';
    inf = __CHILAppend(inf, "The data in this file are at a lower resolution than the raw data, and this parameter is the fractional standard\n"
		       "deviation of the");
    if (~data->ongrid) {
      inf = __CHILAppendWithSpace(inf, numstring);
    }
    inf = __CHILAppendWithSpace(inf, "raw linear reflectivities measured in each output gate and ray, i.e. the standard deviation of Z\n"
				"divided by mean Z.");
  }
  else if (f == ZDR) {
    inf = __CHILAppend(inf, "This parameter is the ratio of the reflectivities measured at horizontal and vertical polarisation.");
  }
  else if (f == LDR) {
    inf = __CHILAppend(inf, "This parameter is the ratio of Zvh to Zh, where Zvh is the power of the horizontally polarised echo returned after\n"
		       "transmission of a vertically polarised pulse, and Zh is the echo power measured and transmitted at horizontal polarisation.");
    if (data->maxLDR > 5.0) {
      inf = __CHILAppend(inf, "Values larger than around -10 dB are usually indicative of ground clutter.");
    }
  }
  else if (f == PHIDP) {
    inf = __CHILAppend(inf, "This parameter is the difference between the phases of the vertically and horizontally polarised return signals.\n"
		       "Positive values are caused by propagation through regions of horizontally aligned hydrometeors.");
  }

  if (!data->raster_info.cloud_data) {
    if (f == ZH) {
      inf = __CHILAppend(inf, "This parameter is the radar reflectivity factor measured at horizontal polarisation.");
    }
    else if (f == V1 || f == WIDTH_H) {
      inf = __CHILAppend(inf, "It was measured at horizontal polarisation.");
    }
    else if (f == V2) {
      inf = __CHILAppend(inf, "It was measured using both horizontally and vertically polarised pulses.");
    }
    if (data->averays * data->avegates > 1) {
      inf = __CHILAppend(inf, "\nIt consists of the linear mean of the raw values recorded at each output gate and ray.");
    }
  }

  if (f == ZH) {

    inf = __CHILAppend(inf, "Calibration convention: in the absence of attenuation, a cloud at 273 K containing one million\n"
		       "100-micron droplets per cubic metre will have a reflectivity of 0 dBZ at all frequencies.");
    if (data->process > 0) {
      inf = __CHILAppend(inf, "Noise has been subtracted using the cloud-free gates at the end of each ray to determine the noise level.");
      if (data->process == 2) {
	inf = __CHILAppend(inf, "The presence of a range-dependent interference in the noise level was taken into account,\n"
			   "but there may still be occasional anomalous horizontal lines in the clouds observed one or two pixels thick.");
      }
    }
    else if (data->polarprocess) {
      inf = __CHILAppend(inf, "Noise has been subtracted assuming a constant noise level for the whole tape.");
    }
    /*
    else {
      inf = __CHILAppend(inf, "Noise subtraction has NOT been performed.");
    }
    */

    if (!data->ray.range_normalised & !data->rangecorrection) {
      inf = __CHILAppend(inf, "The data are NOT range corrected.");
    }

    if (data->clean == 1) {
      inf = __CHILAppend(inf, "To reduce speckle noise, any cloudy pixel that had cloud-free pixels to each side of it in range was removed.");
    }
    else if (data->clean == 2) {
      inf = __CHILAppend(inf, "To reduce speckle noise, any cloudy pixel or pair of cloudy pixels that had cloud-free pixels to each side of them in range were removed.");
    }
  }
  else {
    if (data->polarprocess) {
      inf = __CHILAppend(inf, "Noise has been subtracted assuming a constant noise level for the whole tape.");
    }
    if (f == ZDR) {
      if (data->forceZDRcal) {
	inf = __CHILAppend(inf, "The Zdr calibration figure was specified explicitly in the processing of this file, so should be good to better than 0.5 dB.\n"
			   "The calibration can be checked by making sure that Zdr is on average 0 dB in Rayleigh scattering light drizzle.");
      }
      else {
	inf = __CHILAppend(inf, "The Zdr calibration figure was taken from the original \"sum\" file, so may be incorrect as Zdr can drift with time.\n"
			   "The user is therefore advised to check that Zdr is on average 0 dB in Rayleigh scattering light drizzle.");
      }
    }
  }
  return inf;
}
